// src/mock/mock-client.js
(function() {
  if (window.dvlp) {
    return;
  }
  const RE_WEB_SOCKET_PROTOCOL = /wss?:/;
  const originalXMLHttpRequestOpen = window.XMLHttpRequest.prototype.open;
  const originalFetch = window.fetch;
  const cache = [].map(function(mockData) {
    mockData.originRegex = new RegExp(mockData.originRegex);
    mockData.pathRegex = new RegExp(mockData.pathRegex);
    return mockData;
  });
  const events = cache.reduce(function(events2, mockData) {
    if (mockData.events) {
      events2[mockData.href] = mockData.events;
    }
    return events2;
  }, {});
  let networkDisabled = false;
  let reroute = false;
  window.XMLHttpRequest.prototype.open = function open(method, href) {
    const hrefAndMock = matchHref(href);
    href = hrefAndMock[0];
    const mockData = hrefAndMock[1];
    if (mockData) {
      if (mockData.response) {
        const xhr = this;
        const mockResponse = resolveMockResponse(mockData);
        this.send = function send() {
          if (mockResponse.status === 0) {
            return;
          }
          const body = typeof mockResponse.body === "string" ? mockResponse.body : JSON.stringify(mockResponse.body);
          Object.defineProperties(xhr, {
            readyState: {
              value: 4
            },
            response: {
              get: function() {
                if (mockData.callback) {
                  setTimeout(mockData.callback, 0);
                }
                return body;
              }
            },
            responseText: {
              get: function() {
                return this.response;
              }
            },
            responseURL: {
              value: href
            },
            status: {
              get: function() {
                return mockResponse.status;
              }
            }
          });
          console.log(
            "mocking xhr response (with local data) for: " + parseOriginalHref(href)
          );
          if (xhr.onreadystatechange && typeof xhr.onreadystatechange === "function") {
            xhr.onreadystatechange({ currentTarget: xhr });
          }
          xhr.onload({ currentTarget: xhr });
        };
      } else if (mockData.callback) {
        this.addEventListener("loadend", function() {
          mockData.callback();
        });
      }
    }
    console.log(
      "mocking xhr response (with remote data) for: " + parseOriginalHref(href)
    );
    return originalXMLHttpRequestOpen.call(this, method, href);
  };
  if (typeof Proxy !== "undefined") {
    if (typeof fetch !== "undefined") {
      window.fetch = new Proxy(window.fetch, {
        apply: function(target, ctx, args) {
          const options = args[1] || {};
          const hrefAndMock = matchHref(args[0]);
          const href = hrefAndMock[0];
          const mockData = hrefAndMock[1];
          args[0] = href;
          if (mockData) {
            if (mockData.response) {
              const mockResponse = resolveMockResponse(mockData, options);
              if (mockResponse.status === 0) {
                return new Promise(
                  function() {
                  },
                  function() {
                  }
                );
              }
              const body = typeof mockResponse.body === "string" ? mockResponse.body : JSON.stringify(mockResponse.body);
              const res = new Response(body, {
                headers: mockResponse.headers,
                status: mockResponse.status
              });
              if (mockData.callback) {
                setTimeout(mockData.callback, 0);
              }
              console.log(
                "mocking fetch response (with local data) for: " + parseOriginalHref(href)
              );
              return Promise.resolve(res);
            } else if (mockData.callback) {
              return Reflect.apply(target, ctx, args).then(function(response) {
                setTimeout(mockData.callback, 0);
                return response;
              }).catch(function(err) {
                setTimeout(mockData.callback, 0);
                throw err;
              });
            }
          }
          console.log(
            "mocking fetch response (with remote data) for: " + parseOriginalHref(href)
          );
          return Reflect.apply(target, ctx, args);
        }
      });
    }
    if (typeof EventSource !== "undefined") {
      window.EventSource = new Proxy(window.EventSource, {
        construct: function(target, args) {
          if (!args[0].includes("dvlpreload")) {
            args[0] = matchHref(args[0])[0];
            console.log(
              "mocking EventSource response (with remote data) for: " + args[0]
            );
          }
          return Reflect.construct(target, args);
        }
      });
    }
    if (typeof WebSocket !== "undefined") {
      window.WebSocket = new Proxy(window.WebSocket, {
        construct: function(target, args) {
          args[0] = matchHref(args[0])[0];
          console.log(
            "mocking WebSocket response (with remote data) for: " + args[0]
          );
          return Reflect.construct(target, args);
        }
      });
    }
  }
  window.dvlp = {
    events,
    cache,
    /**
     * Disable all external network connections
     * and optionally reroute all external requests to this server
     *
     * @param { boolean } [rerouteAllRequests]
     * @returns { void }
     */
    disableNetwork: function disableNetwork(rerouteAllRequests) {
      networkDisabled = true;
      reroute = rerouteAllRequests || false;
    },
    /**
     * Re-enable all external network connections
     *
     * @returns { void }
     */
    enableNetwork: function enableNetwork() {
      networkDisabled = false;
      reroute = false;
    },
    /**
     * Add mock response for "req"
     *
     * @param { string | MockRequest } req
     * @param { MockResponse | MockResponseHandler } [res]
     * @param { boolean } [once]
     * @param { () => void } [onMockCallback]
     * @returns { () => void } remove mock instance
     */
    mockResponse: function mockResponse(req, res, once, onMockCallback) {
      const ignoreSearch = req && typeof req === "object" && req.url !== void 0 && req.type === void 0 && req.ignoreSearch || false;
      const url = getUrl(req);
      const originRegex = new RegExp(
        url.origin.replace(/http:|https:/, "https?:").replace("ws:", "wss?:").replace("//", "\\/\\/")
      );
      const pathRegex = new RegExp(url.pathname.replace(/\//g, "\\/"));
      if (typeof res !== "function") {
        if (res && !res.body) {
          res = { body: res, headers: {} };
        }
      }
      const mock = {
        callback: onMockCallback,
        href: url.href,
        ignoreSearch,
        once: once || false,
        originRegex,
        pathRegex,
        response: res,
        search: url.search
      };
      cache.unshift(mock);
      return function() {
        remove(mock);
      };
    },
    /**
     * Trigger EventSource/WebSocket event
     *
     * @param { string } stream
     * @param { string | { message: string | object, options: { event: string, id: string } } } event
     */
    pushEvent: function pushEvent(stream, event) {
      originalFetch("/dvlp/push-event", {
        method: "POST",
        headers: {
          Accept: "application/json",
          "Content-Type": "application/json"
        },
        body: JSON.stringify({ stream, event })
      });
    }
  };
  function matchHref(href) {
    const url = getUrl(href);
    if (url.pathname === "/dvlpreload") {
      return [href];
    }
    const origin = url.origin.indexOf(url.host) === -1 ? url.origin + url.host : url.origin;
    let mockData;
    for (let i = 0; i < cache.length; i++) {
      const mock = cache[i];
      if (!mock.originRegex.test(origin) || !mock.ignoreSearch && mock.search && !isEqualSearch(url.search, mock.search)) {
        continue;
      }
      if (mock.pathRegex.exec(url.pathname) != null) {
        mockData = mock;
        break;
      }
    }
    if (mockData) {
      if (mockData.once) {
        remove(mockData);
      }
      href = (RE_WEB_SOCKET_PROTOCOL.test(url.protocol) ? "ws:" : location.protocol) + "//" + location.host + location.pathname + "?dvlpmock=" + encodeURIComponent(url.href);
    } else if (location.host !== url.host) {
      if (reroute) {
        url.host = location.host;
        href = url.href;
      } else if (networkDisabled) {
        throw Error(
          "network connections disabled. Unable to request " + parseOriginalHref(href)
        );
      }
    }
    return [href, mockData];
  }
  function resolveMockResponse(mockData, requestOptions = {}) {
    const mockResponse = mockData.response;
    const resolved = {
      body: "",
      headers: mockData.response.headers || {},
      status: 0
    };
    if (typeof mockResponse === "function") {
      requestOptions.url = mockData.href;
      mockResponse(requestOptions, {
        end: function end(data) {
          resolved.body = data;
        },
        setHeader: function setHeader(name, value) {
          resolved.headers[name] = value;
        },
        write: function write(chunk) {
          resolved.body += chunk;
        },
        writeHead: function writeHead(statusCode, headers) {
          resolved.status = statusCode;
          if (headers) {
            resolved.headers = headers;
          }
        }
      });
    } else if (mockResponse.error) {
      resolved.body = "error";
      resolved.status = 500;
    } else if (mockResponse.missing) {
      resolved.body = "missing";
      resolved.status = 404;
    } else if (!mockResponse.hang) {
      resolved.body = mockResponse.body;
      resolved.status = mockResponse.status || 200;
    }
    return resolved;
  }
  function getUrl(href) {
    href = typeof href === "string" ? href : href.url;
    return new URL(href, location.href);
  }
  function remove(mockData) {
    for (let i = 0; i < cache.length; i++) {
      if (mockData === cache[i]) {
        cache.splice(i, 1);
      }
    }
  }
  function isEqualSearch(search1, search2) {
    const searchMap1 = parseSearch(search1);
    const searchMap2 = parseSearch(search2);
    if (Object.keys(searchMap1).length !== Object.keys(searchMap2).length) {
      return false;
    }
    for (let key in searchMap1) {
      const values1 = searchMap1[key];
      const values2 = searchMap2[key];
      if (!values2 || values1.length !== values2.length) {
        return false;
      }
      for (let i = 0; i < values1.length; i++) {
        const value = values1[i];
        if (values2.indexOf(value) === -1) {
          return false;
        }
      }
    }
    return true;
  }
  function parseSearch(search) {
    search = search.slice(1).split("&");
    const searchMap = {};
    for (let i = 0; i < search.length; i++) {
      const keyVal = search[i].split("=");
      const key = keyVal[0];
      const val = keyVal[1];
      if (!(key in searchMap)) {
        searchMap[key] = [];
      }
      searchMap[key].push(val);
    }
    return searchMap;
  }
  function parseOriginalHref(hrefOrRequest) {
    const href = typeof hrefOrRequest === "string" ? hrefOrRequest : hrefOrRequest.url;
    if (href.indexOf("?dvlpmock") === -1) {
      return href;
    }
    return decodeURIComponent(href.split("?dvlpmock=")[1]);
  }
})();

// src/test-browser/index.js
var testBrowser = {
  /**
   * Disable all external network connections
   * and optionally reroute all external requests to this server
   *
   * @param { boolean } [rerouteAllRequests]
   */
  disableNetwork(rerouteAllRequests) {
    return window.dvlp.disableNetwork(rerouteAllRequests);
  },
  /**
   * Re-enable all external network connections
   */
  enableNetwork() {
    return window.dvlp.enableNetwork();
  },
  /**
   * Add mock response for "req"
   *
   * @param { string | MockRequest } req
   * @param { MockResponse } [res]
   * @param { boolean } [once]
   * @param { () => void } [onMockCallback]
   * @returns { () => void } remove mock instance
   */
  mockResponse(req, res, once, onMockCallback) {
    return window.dvlp.mockResponse(req, res, once, onMockCallback);
  },
  /**
   * Trigger EventSource/WebSocket event
   *
   * @param { string } stream
   * @param { string | { message: string | object, options: { event: string, id: string } } } event
   */
  pushEvent(stream, event) {
    return window.dvlp.pushEvent(stream, event);
  }
};
export {
  testBrowser
};
